@@ -1,8 +1,6 @@
 # Copyright: Damien Elmes <anki@ichi2.net>
 # License: GNU AGPL, version 3 or later; http://www.gnu.org/licenses/agpl.html
 
-from __future__ import division
-import socket
 import time
 import traceback
 import gc
@@ -45,14 +43,22 @@ class SyncManager(QObject):
         # create the thread, setup signals and start running
         t = self.thread = SyncThread(
             self.pm.collectionPath(), self.pm.profile['syncKey'],
-            auth=auth, media=self.pm.profile['syncMedia'])
-        self.connect(t, SIGNAL("event"), self.onEvent)
+            auth=auth, media=self.pm.profile['syncMedia'],
+            hostNum=self.pm.profile.get("hostNum"),
+        )
+        t.event.connect(self.onEvent)
         self.label = _("Connecting...")
-        self.mw.progress.start(immediate=True, label=self.label)
+        prog = self.mw.progress.start(immediate=True, label=self.label)
         self.sentBytes = self.recvBytes = 0
         self._updateLabel()
         self.thread.start()
         while not self.thread.isFinished():
+            if prog.wantCancel:
+                self.thread.flagAbort()
+                # make sure we don't display 'upload success' msg
+                self._didFullUp = False
+                # abort may take a while
+                self.mw.progress.update(_("Stopping..."))
             self.mw.app.processEvents()
             self.thread.wait(100)
         self.mw.progress.finish()
@@ -60,6 +66,7 @@ class SyncManager(QObject):
             showText(self.thread.syncMsg)
         if self.thread.uname:
             self.pm.profile['syncUser'] = self.thread.uname
+        self.pm.profile['hostNum'] = self.thread.hostNum
         def delayedInfo():
             if self._didFullUp and not self._didError:
                 showInfo(_("""\
@@ -74,9 +81,9 @@ automatically."""))
     def _updateLabel(self):
         self.mw.progress.update(label="%s\n%s" % (
             self.label,
-            _("%(a)dkB up, %(b)dkB down") % dict(
-                a=self.sentBytes // 1024,
-                b=self.recvBytes // 1024)))
+            _("%(a)0.1fkB up, %(b)0.1fkB down") % dict(
+                a=self.sentBytes / 1024,
+                b=self.recvBytes / 1024)))
 
     def onEvent(self, evt, *args):
         pu = self.mw.progress.update
@@ -109,7 +116,7 @@ automatically."""))
             elif t == "sanity":
                 m = _("Checking...")
             elif t == "findMedia":
-                m = _("Syncing Media...")
+                m = _("Checking media...")
             elif t == "upgradeRequired":
                 showText(_("""\
 Please visit AnkiWeb, upgrade your deck, then try again."""))
@@ -135,12 +142,14 @@ sync again to correct the issue."""))
             pass
         elif evt == "fullSync":
             self._confirmFullSync()
+        elif evt == "downloadClobber":
+            showInfo(_("Your AnkiWeb collection does not contain any cards. Please sync again and choose 'Upload' instead."))
         elif evt == "send":
             # posted events not guaranteed to arrive in order
-            self.sentBytes = max(self.sentBytes, args[0])
+            self.sentBytes = max(self.sentBytes, int(args[0]))
             self._updateLabel()
         elif evt == "recv":
-            self.recvBytes = max(self.recvBytes, args[0])
+            self.recvBytes = max(self.recvBytes, int(args[0]))
             self._updateLabel()
 
     def _rewriteError(self, err):
@@ -175,7 +184,7 @@ AnkiWeb is too busy at the moment. Please try again in a few minutes.""")
                 "Antivirus or firewall software is preventing Anki from connecting to the internet.")
         elif "10054" in err or "Broken pipe" in err:
             return _("Connection timed out. Either your internet connection is experiencing problems, or you have a very large file in your media folder.")
-        elif "Unable to find the server" in err:
+        elif "Unable to find the server" in err or "socket.gaierror" in err:
             return _(
                 "Server not found. Either your connection is down, or antivirus/firewall "
                 "software is blocking Anki from connecting to the internet.")
@@ -184,9 +193,9 @@ AnkiWeb is too busy at the moment. Please try again in a few minutes.""")
         elif "code: 413" in err:
             return _("Your collection or a media file is too large to sync.")
         elif "EOF occurred in violation of protocol" in err:
-            return _("Error establishing a secure connection. This is usually caused by antivirus, firewall or VPN software, or problems with your ISP.")
+            return _("Error establishing a secure connection. This is usually caused by antivirus, firewall or VPN software, or problems with your ISP.") + " (eof)"
         elif "certificate verify failed" in err:
-            return _("Error establishing a secure connection. This is usually caused by antivirus, firewall or VPN software, or problems with your ISP.")
+            return _("Error establishing a secure connection. This is usually caused by antivirus, firewall or VPN software, or problems with your ISP.") + " (invalid cert)"
         return err
 
     def _getUserPass(self):
@@ -217,8 +226,8 @@ enter your details below.""") %
         vbox.addLayout(g)
         bb = QDialogButtonBox(QDialogButtonBox.Ok|QDialogButtonBox.Cancel)
         bb.button(QDialogButtonBox.Ok).setAutoDefault(True)
-        self.connect(bb, SIGNAL("accepted()"), d.accept)
-        self.connect(bb, SIGNAL("rejected()"), d.reject)
+        bb.accepted.connect(d.accept)
+        bb.rejected.connect(d.reject)
         vbox.addWidget(bb)
         d.setLayout(vbox)
         d.show()
@@ -230,7 +239,14 @@ enter your details below.""") %
         return (u, p)
 
     def _confirmFullSync(self):
-        diag = askUserDialog(_("""\
+        self.mw.progress.finish()
+        if self.thread.localIsEmpty:
+            diag = askUserDialog(
+                _("Local collection has no cards. Download from AnkiWeb?"),
+                [_("Download from AnkiWeb"), _("Cancel")])
+            diag.setDefault(1)
+        else:
+            diag = askUserDialog(_("""\
 Your decks here and on AnkiWeb differ in such a way that they can't \
 be merged together, so it's necessary to overwrite the decks on one \
 side with the decks from the other.
@@ -248,7 +264,7 @@ automatically."""),
                 [_("Upload to AnkiWeb"),
                  _("Download from AnkiWeb"),
                  _("Cancel")])
-        diag.setDefault(2)
+            diag.setDefault(2)
         ret = diag.run()
         if ret == _("Upload to AnkiWeb"):
             self.thread.fullSyncChoice = "upload"
@@ -256,6 +272,7 @@ automatically."""),
             self.thread.fullSyncChoice = "download"
         else:
             self.thread.fullSyncChoice = "cancel"
+        self.mw.progress.start(immediate=True)
 
     def _clockOff(self):
         showWarning(_("""\
@@ -276,12 +293,19 @@ Check Database, then sync again."""))
 
 class SyncThread(QThread):
 
-    def __init__(self, path, hkey, auth=None, media=True):
+    event = pyqtSignal(str, str)
+
+    def __init__(self, path, hkey, auth=None, media=True, hostNum=None):
         QThread.__init__(self)
         self.path = path
         self.hkey = hkey
         self.auth = auth
         self.media = media
+        self.hostNum = hostNum
+        self._abort = 0 # 1=flagged, 2=aborting
+
+    def flagAbort(self):
+        self._abort = 1
 
     def run(self):
         # init this first so an early crash doesn't cause an error
@@ -293,28 +317,28 @@ class SyncThread(QThread):
         except:
             self.fireEvent("corrupt")
             return
-        self.server = RemoteServer(self.hkey)
+        self.server = RemoteServer(self.hkey, hostNum=self.hostNum)
         self.client = Syncer(self.col, self.server)
         self.sentTotal = 0
         self.recvTotal = 0
-        # throttle updates; qt doesn't handle lots of posted events well
-        self.byteUpdate = time.time()
         def syncEvent(type):
             self.fireEvent("sync", type)
         def syncMsg(msg):
             self.fireEvent("syncMsg", msg)
-        def canPost():
-            if (time.time() - self.byteUpdate) > 0.1:
-                self.byteUpdate = time.time()
-                return True
         def sendEvent(bytes):
-            self.sentTotal += bytes
-            if canPost():
-                self.fireEvent("send", self.sentTotal)
+            if not self._abort:
+                self.sentTotal += bytes
+                self.fireEvent("send", str(self.sentTotal))
+            elif self._abort == 1:
+                self._abort = 2
+                raise Exception("sync cancelled")
         def recvEvent(bytes):
-            self.recvTotal += bytes
-            if canPost():
-                self.fireEvent("recv", self.recvTotal)
+            if not self._abort:
+                self.recvTotal += bytes
+                self.fireEvent("recv", str(self.recvTotal))
+            elif self._abort == 1:
+                self._abort = 2
+                raise Exception("sync cancelled")
         addHook("sync", syncEvent)
         addHook("syncMsg", syncMsg)
         addHook("httpSend", sendEvent)
@@ -324,8 +348,6 @@ class SyncThread(QThread):
             self._sync()
         except:
             err = traceback.format_exc()
-            if not isinstance(err, unicode):
-                err = unicode(err, "utf8", "replace")
             self.fireEvent("error", err)
         finally:
             # don't bump mod time unless we explicitly save
@@ -335,6 +357,16 @@ class SyncThread(QThread):
             remHook("httpSend", sendEvent)
             remHook("httpRecv", recvEvent)
 
+    def _abortingSync(self):
+        try:
+            return self.client.sync()
+        except Exception as e:
+            if "sync cancelled" in str(e):
+                self.server.abort()
+                raise
+            else:
+                raise
+
     def _sync(self):
         if self.auth:
             # need to authenticate and obtain host key
@@ -347,19 +379,17 @@ class SyncThread(QThread):
                 self.fireEvent("newKey", self.hkey)
         # run sync and check state
         try:
-            ret = self.client.sync()
-        except Exception, e:
+            ret = self._abortingSync()
+        except Exception as e:
             log = traceback.format_exc()
             err = repr(str(e))
             if ("Unable to find the server" in err or
-                "Errno 2" in err):
+                "Errno 2" in err or "getaddrinfo" in err):
                 self.fireEvent("offline")
+            elif "sync cancelled" in err:
+                pass
             else:
-                if not err:
-                    err = log
-                if not isinstance(err, unicode):
-                    err = unicode(err, "utf8", "replace")
-                self.fireEvent("error", err)
+                self.fireEvent("error", log)
             return
         if ret == "badAuth":
             return self.fireEvent("badAuth")
@@ -381,29 +411,36 @@ class SyncThread(QThread):
             self.fireEvent("error", "Unknown sync return code.")
         self.syncMsg = self.client.syncMsg
         self.uname = self.client.uname
+        self.hostNum = self.client.hostNum
         # then move on to media sync
         self._syncMedia()
 
     def _fullSync(self):
-        # if the local deck is empty, assume user is trying to download
-        if self.col.isEmpty():
-            f = "download"
-        else:
-            # tell the calling thread we need a decision on sync direction, and
-            # wait for a reply
-            self.fullSyncChoice = False
-            self.fireEvent("fullSync")
-            while not self.fullSyncChoice:
-                time.sleep(0.1)
-            f = self.fullSyncChoice
+        # tell the calling thread we need a decision on sync direction, and
+        # wait for a reply
+        self.fullSyncChoice = False
+        self.localIsEmpty = self.col.isEmpty()
+        self.fireEvent("fullSync")
+        while not self.fullSyncChoice:
+            time.sleep(0.1)
+        f = self.fullSyncChoice
         if f == "cancel":
             return
-        self.client = FullSyncer(self.col, self.hkey, self.server.con)
-        if f == "upload":
-            if not self.client.upload():
-                self.fireEvent("upbad")
-        else:
-            self.client.download()
+        self.client = FullSyncer(self.col, self.hkey, self.server.client,
+                                 hostNum=self.hostNum)
+        try:
+            if f == "upload":
+                if not self.client.upload():
+                    self.fireEvent("upbad")
+            else:
+                ret = self.client.download()
+                if ret == "downloadClobber":
+                    self.fireEvent(ret)
+                    return
+        except Exception as e:
+            if "sync cancelled" in str(e):
+                return
+            raise
         # reopen db and move on to media sync
         self.col.reopen()
         self._syncMedia()
@@ -411,101 +448,23 @@ class SyncThread(QThread):
     def _syncMedia(self):
         if not self.media:
             return
-        self.server = RemoteMediaServer(self.col, self.hkey, self.server.con)
+        self.server = RemoteMediaServer(self.col, self.hkey, self.server.client,
+                                        hostNum=self.hostNum)
         self.client = MediaSyncer(self.col, self.server)
-        ret = self.client.sync()
+        try:
+            ret = self.client.sync()
+        except Exception as e:
+            if "sync cancelled" in str(e):
+                return
+            raise
         if ret == "noChanges":
             self.fireEvent("noMediaChanges")
-        elif ret == "sanityCheckFailed":
+        elif ret == "sanityCheckFailed" or ret == "corruptMediaDB":
             self.fireEvent("mediaSanity")
         else:
             self.fireEvent("mediaSuccess")
 
-    def fireEvent(self, *args):
-        self.emit(SIGNAL("event"), *args)
+    def fireEvent(self, cmd, arg=""):
+        self.event.emit(cmd, arg)
 
 
-# Monkey-patch httplib & httplib2 so we can get progress info
-######################################################################
-
-CHUNK_SIZE = 65536
-import httplib, httplib2
-from cStringIO import StringIO
-from anki.hooks import runHook
-
-# sending in httplib
-def _incrementalSend(self, data):
-    """Send `data' to the server."""
-    if self.sock is None:
-        if self.auto_open:
-            self.connect()
-        else:
-            raise httplib.NotConnected()
-    # if it's not a file object, make it one
-    if not hasattr(data, 'read'):
-        if isinstance(data, unicode):
-            data = data.encode("utf8")
-        data = StringIO(data)
-    while 1:
-        block = data.read(CHUNK_SIZE)
-        if not block:
-            break
-        self.sock.sendall(block)
-        runHook("httpSend", len(block))
-
-httplib.HTTPConnection.send = _incrementalSend
-
-# receiving in httplib2
-# this is an augmented version of httplib's request routine that:
-# - doesn't assume requests will be tried more than once
-# - calls a hook for each chunk of data so we can update the gui
-# - retries only when keep-alive connection is closed
-def _conn_request(self, conn, request_uri, method, body, headers):
-    for i in range(2):
-        try:
-            if conn.sock is None:
-              conn.connect()
-            conn.request(method, request_uri, body, headers)
-        except socket.timeout:
-            raise
-        except socket.gaierror:
-            conn.close()
-            raise httplib2.ServerNotFoundError(
-                "Unable to find the server at %s" % conn.host)
-        except httplib2.ssl_SSLError:
-            conn.close()
-            raise
-        except socket.error, e:
-            conn.close()
-            raise
-        except httplib.HTTPException:
-            conn.close()
-            raise
-        try:
-            response = conn.getresponse()
-        except httplib.BadStatusLine:
-            print "retry bad line"
-            conn.close()
-            conn.connect()
-            continue
-        except (socket.error, httplib.HTTPException):
-            raise
-        else:
-            content = ""
-            if method == "HEAD":
-                response.close()
-            else:
-                buf = StringIO()
-                while 1:
-                    data = response.read(CHUNK_SIZE)
-                    if not data:
-                        break
-                    buf.write(data)
-                    runHook("httpRecv", len(data))
-                content = buf.getvalue()
-            response = httplib2.Response(response)
-            if method != "HEAD":
-                content = httplib2._decompressContent(response, content)
-        return (response, content)
-
-httplib2.Http._conn_request = _conn_request
